import math
from dataclasses import dataclass, field
from typing import List, Literal, Mapping, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from entity_gym.env import ObsSpace
from ragged_buffer import RaggedBufferI64

from rogue_net.input_norm import InputNorm


@dataclass(frozen=True, eq=False)
class RelposEncodingConfig:
    """Settings for relative position encoding.

    Attributes:
        extent: Each integer relative position in the interval [-extent, extent] receives a positional embedding, with positions outside the interval snapped to the closest end.
        position_features: Names of position features used for relative position encoding.
        scale: Relative positions are divided by the scale before being assigned an embedding.
        per_entity_values: Whether to use per-entity embeddings for relative positional values.
        exclude_entities: List of entity types to exclude from relative position encoding.
        key_relpos_projection: Adds a learnable projection from the relative position/distance to the relative positional keys.
        value_relpos_projection: Adds a learnable projection from the relative position/distance to the relative positional values.
        per_entity_projections: Uses a different learned projection per entity type for the `key_relpos_projection` and `value_relpos_projection`.
        radial: Buckets all relative positions by their angle. The `extent` is interpreted as the number of buckets.
        distance: Buckets all relative positions by their distance. The `extent` is interpreted as the number of buckets.
        rotation_vec_features: Name of features that give a unit orientation vector for each entity by which to rotate relative positions.
        rotation_angle_feature: Name of feature that gives an angle in radians by which to rotate relative positions.
        interpolate: Whether to interpolate between the embeddings of neighboring positions.
    """

    extent: List[int]
    position_features: List[str]
    scale: float = 1.0
    per_entity_values: bool = False
    exclude_entities: List[str] = field(default_factory=list)
    value_relpos_projection: bool = False
    key_relpos_projection: bool = False
    per_entity_projections: bool = False
    radial: bool = False
    distance: bool = False
    rotation_vec_features: Optional[List[str]] = None
    rotation_angle_feature: Optional[str] = None
    interpolate: bool = False
    value_gate: Literal["linear", "relu", "gelu", "sigmoid", None] = "relu"

    def __post_init__(self) -> None:
        if self.radial and self.distance:
            assert (
                len(self.extent) == 2
            ), "Polar relative position encoding expects two extent values (number of angle buckets and number of distance buckets)"
        elif self.radial:
            assert (
                len(self.extent) == 1
            ), "Radial relative position encoding expects a single extent value (number of angle buckets)"
            assert (
                self.rotation_angle_feature is not None
                or self.rotation_vec_features is not None
            ), "Radial relative position encoding requires `rotation_angle_feature` or `rotation_vec_features` to be set"
        elif self.distance:
            assert (
                len(self.extent) == 1
            ), "Distance relative position encoding expects a single extent value (number of distance buckets)"
        else:
            assert len(self.extent) == len(
                self.position_features
            ), "Relative position encoding expects a extent value for each position feature"
        assert (
            self.rotation_vec_features is None or self.rotation_angle_feature is None
        ), "Only one of rotation_vec_features and rotation_angle_feature can be specified"


class RelposEncoding(nn.Module, RelposEncodingConfig):
    def __init__(
        self, config: RelposEncodingConfig, obs_space: ObsSpace, dmodel: int, dhead: int
    ) -> None:
        nn.Module.__init__(self)
        RelposEncodingConfig.__init__(self, **config.__dict__)

        self.n_entity = len(obs_space.entities)
        if self.radial and self.distance:
            angles, distances = self.extent
            strides = [1.0, angles]
            positions = angles * distances
        elif self.radial or self.distance:
            strides = [1.0]
            positions = self.extent[0]
        else:
            strides = []
            positions = 1
            for extent in self.extent:
                strides.append(float(positions))
                positions *= 2 * extent + 1
        self.positions = positions
        self.register_buffer("strides", torch.tensor(strides).unsqueeze(0))
        self.register_buffer(
            "extent_tensor", torch.tensor(self.extent).view(1, 1, 1, -1).long()
        )
        # TODO: tune embedding init scale
        self.keys = nn.Embedding(self.positions, dhead)
        self.values = nn.Embedding(
            self.positions * self.n_entity
            if config.per_entity_values
            else self.positions,
            dhead,
        )
        self.distance_values = nn.Embedding(self.n_entity, dhead)
        self.keys.weight.data.normal_(mean=0.0, std=0.05)
        self.values.weight.data.normal_(mean=0.0, std=0.2)
        self.distance_values.weight.data.normal_(mean=0.0, std=0.2)
        self.position_feature_indices = {
            entity_name: torch.LongTensor(
                [
                    entity.features.index(feature_name)
                    for feature_name in config.position_features
                ]
            )
            for entity_name, entity in obs_space.entities.items()
            if entity_name not in self.exclude_entities
        }
        self.orientation_vec_indices = (
            {
                entity_name: torch.LongTensor(
                    [
                        entity.features.index(feature_name)
                        for feature_name in self.rotation_vec_features
                    ]
                )
                for entity_name, entity in obs_space.entities.items()
                if entity_name not in self.exclude_entities
                and all(
                    feature in entity.features for feature in self.rotation_vec_features
                )
            }
            if self.rotation_vec_features is not None
            else None
        )
        self.orientation_angle_index = (
            {
                entity_name: entity.features.index(self.rotation_angle_feature)
                for entity_name, entity in obs_space.entities.items()
                if entity_name not in self.exclude_entities
                and self.rotation_angle_feature in entity.features
            }
            if self.rotation_angle_feature is not None
            else None
        )
        if self.value_relpos_projection:
            if self.per_entity_projections:
                self.vproj: nn.Module = nn.ModuleDict(
                    {
                        entity_name: nn.Linear(3, dhead)
                        for entity_name in self.position_feature_indices
                    }
                )
            else:
                self.vproj = nn.Linear(3, dhead)
        if self.key_relpos_projection:
            if self.per_entity_projections:
                self.kproj: nn.Module = nn.ModuleDict(
                    {
                        entity_name: nn.Linear(3, dhead)
                        for entity_name in self.position_feature_indices
                    }
                )
            else:
                self.kproj = nn.Linear(3, dhead)
        if self.key_relpos_projection or self.value_relpos_projection:
            self.relpos_norm = InputNorm(3)
        if self.value_gate is not None:
            self.value_gate_proj = nn.Linear(dmodel, dhead)
        self.cached_rkvs: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
        self.global_entity = "__global__" in obs_space.entities

    def relattn_logits(self, queries: torch.Tensor) -> torch.Tensor:
        assert self.cached_rkvs is not None
        relkeys = self.cached_rkvs[0]  #       (B, T, T, dhead)
        # Broadcast and sum over last dimension (dot product of queries with relative keys)
        relattn: torch.Tensor = torch.einsum("bhsd,bstd->bhst", queries, relkeys) * (
            1.0 / math.sqrt(relkeys.size(-1))
        )  # (B, nh, T, T)
        return relattn

    def relpos_values(self, att: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        assert self.cached_rkvs is not None
        relvals = self.cached_rkvs[1]  #       (B, T_query, T_target, dhead)
        if self.value_gate is not None:
            vgate = self.value_gate_proj(x)  # (B, T_target, dhead)
            if self.value_gate == "relu":
                vgate = F.relu(vgate)
            elif self.value_gate == "gelu":
                vgate = F.gelu(vgate)
            elif self.value_gate == "sigmoid":
                vgate = torch.sigmoid(vgate)

            relvals = torch.einsum("bqtd,btd->bqtd", relvals, vgate)
        rely: torch.Tensor = torch.einsum(
            "bhst,bstd->bhsd", att, relvals
        )  # (B, nh, T, T)
        return rely

    def keys_values(
        self,
        # Dict from entity name to raw input features
        x: Mapping[str, torch.Tensor],
        # Maps entities ordered first by type to entities ordered first by frame
        index_map: torch.Tensor,
        # Maps flattened embeddings to packed/padded tensor with fixed sequence lengths
        packpad_index: Optional[torch.Tensor],
        # Ragged shape of the flattened embeddings tensor
        shape: RaggedBufferI64,
        # Type of each entity
        entity_type: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if "__global__" in x and not self.global_entity:
            x = {k: v for k, v in x.items() if k != "__global__"}
        relative_positions, entity_type = self._relative_positions(
            x, index_map, packpad_index, shape, entity_type
        )
        torientation = (
            self._orientations(x, index_map, packpad_index, shape)
            if self.radial
            else None
        )

        if self.interpolate:
            indices_weights = self._interpolated_partition(
                relative_positions, torientation
            )

            keys: torch.Tensor = sum(  # type: ignore
                self.keys(indices) * weights.unsqueeze(-1)
                for indices, weights in indices_weights
            )
            if self.per_entity_values:
                per_entity_indices_weights = [
                    (
                        indices
                        + (entity_type * self.positions)
                        .transpose(2, 1)
                        .long()
                        .repeat(1, indices.size(2), 1),
                        weights,
                    )
                    for indices, weights in indices_weights
                ]
            else:
                per_entity_indices_weights = indices_weights
            values: torch.Tensor = sum(  # type: ignore
                self.values(indices) * weights.unsqueeze(-1)
                for indices, weights in per_entity_indices_weights
            )
        else:
            indices = self._partition(relative_positions, torientation)
            # Batch x Seq x Seq x d_model
            keys = self.keys(indices)
            if self.per_entity_values:
                per_entity_type_indices = indices + (
                    entity_type * self.positions
                ).transpose(2, 1).long().repeat(1, indices.size(2), 1)
            else:
                per_entity_type_indices = indices
            values = self.values(per_entity_type_indices)

        if self.value_relpos_projection or self.key_relpos_projection:
            # TODO: torch.norm deprecated, does this do the right thing?
            dist = relative_positions.norm(p=2, dim=-1).unsqueeze(-1)
            relpos_dist = torch.cat([relative_positions, dist], dim=-1)
            norm_relpos_dist = self.relpos_norm(relpos_dist)
            if self.per_entity_projections:
                # TODO: efficiency
                if self.value_relpos_projection:
                    for i, vproj in enumerate(self.vproj.values()):  # type: ignore
                        v = vproj(norm_relpos_dist)
                        v[entity_type.squeeze(-1) != 0, :, :] = 0.0
                        values += v
                if self.key_relpos_projection:
                    for i, kproj in enumerate(self.kproj.values()):  # type: ignore
                        k = kproj(norm_relpos_dist)
                        k[entity_type.squeeze(-1) != 0, :, :] = 0.0
                        keys += k
            else:
                if self.value_relpos_projection:
                    values += self.vproj(norm_relpos_dist)
                if self.key_relpos_projection:
                    keys += self.kproj(norm_relpos_dist)

        return keys, values

    def _relative_positions(
        self,
        x: Mapping[str, torch.Tensor],
        index_map: torch.Tensor,
        packpad_index: Optional[torch.Tensor],
        shape: RaggedBufferI64,
        entity_type: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        positions = []
        for entity_name, features in x.items():
            if entity_name not in self.exclude_entities:
                positions.append(
                    features[:, self.position_feature_indices[entity_name]]
                )
            else:
                # TODO: add padding or something?
                raise NotImplementedError(
                    "exclude_entities not implemented for relative position encoding"
                )
        # Flat tensor of positions
        tpos = torch.cat(positions, dim=0)
        # Flat tensor of positions ordered by sample
        tpos = tpos[index_map]
        # Padded/packed Batch x Seq x Pos tensor of positions
        if packpad_index is not None:
            tpos = tpos[packpad_index]
            entity_type = entity_type[packpad_index]
        else:
            size1 = shape.size1(0) if len(tpos) > 0 else 0
            tpos = tpos.reshape(shape.size0(), size1, 2)
            entity_type = entity_type.reshape(shape.size0(), size1, 1)
        # Batch x Seq(q) x Seq(k) x Pos relative positions
        return tpos.unsqueeze(2) - tpos.unsqueeze(1), entity_type

    def _orientations(
        self,
        x: Mapping[str, torch.Tensor],
        index_map: torch.Tensor,
        packpad_index: Optional[torch.Tensor],
        shape: RaggedBufferI64,
    ) -> torch.Tensor:
        """
        Maps a sequence of relative positions to indices.
        """
        # Get entity orientations
        orientations = []
        if self.orientation_angle_index is not None:
            for entity_name, feature in x.items():
                feature_index = self.orientation_angle_index.get(entity_name)
                if feature_index is None:
                    orientations.append(
                        torch.zeros_like(feature[:, 0], device=index_map.device)
                    )
                else:
                    orientations.append(feature[:, feature_index])
            torientation = torch.cat(orientations, dim=0)
        elif self.orientation_vec_indices is not None:
            for entity_name, feature in x.items():
                feature_indices = self.orientation_vec_indices.get(entity_name)
                if feature_indices is None:
                    orientations.append(
                        torch.zeros_like(feature[:, 0], device=index_map.device)
                    )
                else:
                    orientation_vec = feature[:, feature_indices]
                    orientation = torch.atan2(
                        orientation_vec[:, 1], orientation_vec[:, 0]
                    )
                    orientations.append(orientation)
            torientation = torch.cat(orientations, dim=0)[:, None]
        else:
            raise ValueError("No orientation information")

        torientation = torientation[index_map]
        if packpad_index is not None:
            return torientation[packpad_index].view(
                packpad_index.size(0), packpad_index.size(1), 1
            )
        else:
            size1 = shape.size1(0) if shape.items() > 0 else 0
            return torientation.reshape(shape.size0(), size1, 1)

    def _partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(q) x Seq(k) x Pos relative positions
        torientation: Optional[torch.Tensor],
    ) -> torch.Tensor:
        """
        Maps a sequence of relative positions to indices.
        """
        if self.radial and self.distance:
            return self._polar_partition(relative_positions, torientation)
        elif self.radial:
            return self._radial_partition(relative_positions, torientation)
        elif self.distance:
            return self._distance_partition(relative_positions)
        else:
            return self._grid_partition(relative_positions, torientation)

    def _interpolated_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(q) x Seq(k) x Pos relative positions
        torientation: Optional[torch.Tensor],
    ) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        """
        Maps a sequence of relative positions to indices.
        """
        if self.radial and self.distance:
            return self._interpolated_polar_partition(relative_positions, torientation)
        if self.radial:
            return self._interpolated_radial_partition(relative_positions, torientation)
        elif self.distance:
            return self._interpolated_distance_partition(relative_positions)
        else:
            raise NotImplementedError("interpolated grid partition not implemented")

    def _grid_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(k) x Seq(q) x Pos relative positions
        torientation: Optional[torch.Tensor],
    ) -> torch.Tensor:
        assert torientation is None, "Not implemented for non-radial"
        clamped_positions = torch.max(
            torch.min(
                self.extent_tensor,  # type: ignore
                relative_positions * (1.0 / self.scale),
            ),
            -self.extent_tensor,  # type: ignore
        )
        positive_positions = clamped_positions + self.extent_tensor
        return (positive_positions * self.strides).sum(dim=-1).round().long()

    def _radial_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(k) x Seq(q) x Pos relative positions
        torientation: Optional[torch.Tensor],
    ) -> torch.Tensor:
        angles = torch.atan2(
            relative_positions[:, :, :, 1], relative_positions[:, :, :, 0]
        )
        # We need to be careful about ensuring that we don't create indices that fall outside of the extent.
        # Specifically, taking the modulo of a small negative number can round up to the modulus:
        # (torch.tensor([-1e-12], dtype=torch.float32) % torch.tensor([2]).long()).long() == torch.tensor([2]).long()
        # We can avoid this by ensuring that `angles` is always positive.
        if torientation is not None:
            angles = angles - torientation + 2 * math.pi
        return (angles / (2 * math.pi) * self.extent[0] % self.extent[0]).long()

    def _interpolated_radial_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(k) x Seq(q) x Pos relative positions
        torientation: Optional[torch.Tensor],
    ) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        angles = torch.atan2(
            relative_positions[:, :, :, 1], relative_positions[:, :, :, 0]
        )
        if torientation is not None:
            angles = angles - torientation + 2 * math.pi
        norm_angles = angles / (2 * math.pi) * self.extent[0] % self.extent[0]
        index1 = norm_angles.long()
        index2 = (index1 + 1) % self.extent[0]
        weight1 = (index2 - norm_angles) % 1
        weight2 = 1.0 - weight1
        return [(index1, weight1), (index2, weight2)]

    def _distance_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(k) x Seq(q) x Pos relative positions
    ) -> torch.Tensor:
        distances: torch.Tensor = torch.linalg.norm(relative_positions, dim=-1)
        return torch.min(
            (distances * (1.0 / self.scale)).long(),
            # For polar relative positions, distance extent is last element.
            self.extent_tensor[0, 0, 0, -1] - 1,  # type: ignore
        )

    def _interpolated_distance_partition(
        self,
        relative_positions: torch.Tensor,  # Batch x Seq(k) x Seq(q) x Pos relative positions
    ) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        distances: torch.Tensor = torch.linalg.norm(relative_positions, dim=-1) * (
            1.0 / self.scale
        )
        index1 = torch.min(
            distances.long(),
            self.extent_tensor[0, 0, 0, -1] - 1,  # type: ignore
        )
        index2 = torch.min(
            index1 + 1,
            self.extent_tensor[0, 0, 0, -1] - 1,  # type: ignore
        )
        weight1 = index2 - distances
        weight2 = 1.0 - weight1
        return [(index1, weight1), (index2, weight2)]

    def _polar_partition(
        self,
        relative_positions: torch.Tensor,
        torientation: Optional[torch.Tensor],
    ) -> torch.Tensor:
        aindices = self._radial_partition(relative_positions, torientation)
        dindices = self._distance_partition(relative_positions)
        indices = (torch.stack([aindices, dindices], dim=-1) * self.strides.long()).sum(  # type: ignore
            dim=-1
        )
        return indices

    def _interpolated_polar_partition(
        self,
        relative_positions: torch.Tensor,
        torientation: Optional[torch.Tensor],
    ) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        (aindex1, aweight1), (aindex2, aweight2) = self._interpolated_radial_partition(
            relative_positions, torientation
        )
        (dindex1, dweight1), (
            dindex2,
            dweight2,
        ) = self._interpolated_distance_partition(relative_positions)
        indices1 = (torch.stack([aindex1, dindex1], dim=-1) * self.strides.long()).sum(dim=-1)  # type: ignore
        weights1 = aweight1 * dweight1
        indices2 = (torch.stack([aindex2, dindex1], dim=-1) * self.strides.long()).sum(dim=-1)  # type: ignore
        weights2 = aweight2 * dweight1
        indices3 = (torch.stack([aindex1, dindex2], dim=-1) * self.strides.long()).sum(dim=-1)  # type: ignore
        weights3 = aweight1 * dweight2
        indices4 = (torch.stack([aindex2, dindex2], dim=-1) * self.strides.long()).sum(dim=-1)  # type: ignore
        weights4 = aweight2 * dweight2
        return [
            (indices1, weights1),
            (indices2, weights2),
            (indices3, weights3),
            (indices4, weights4),
        ]



from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
import torch
from entity_gym.env import VecActionMask, VecCategoricalActionMask
from ragged_buffer import RaggedBufferI64
from torch import nn
from torch.distributions.categorical import Categorical

from rogue_net.ragged_tensor import RaggedTensor


class CategoricalActionHead(nn.Module):
    def __init__(self, d_model: int, n_choice: int) -> None:
        super().__init__()
        self.d_model = d_model
        self.n_choice = n_choice
        self.proj = layer_init(nn.Linear(d_model, n_choice), std=0.01)

    def forward(
        self,
        x: RaggedTensor,
        index_offsets: RaggedBufferI64,
        mask: VecActionMask,
        prev_actions: Optional[RaggedBufferI64],
    ) -> Tuple[
        torch.Tensor, npt.NDArray[np.int64], torch.Tensor, torch.Tensor, torch.Tensor
    ]:
        assert isinstance(
            mask, VecCategoricalActionMask
        ), f"Expected CategoricalActionMaskBatch, got {type(mask)}"

        device = x.data.device
        lengths = mask.actors.size1()
        if len(mask.actors) == 0:
            return (
                torch.zeros((0), dtype=torch.int64, device=device),
                lengths,
                torch.zeros((0), dtype=torch.float32, device=device),
                torch.zeros((0), dtype=torch.float32, device=device),
                torch.zeros((0, self.n_choice), dtype=torch.float32, device=device),
            )

        actors = (
            torch.tensor((mask.actors + index_offsets).as_array())
            .to(x.data.device)
            .squeeze(-1)
        )
        actor_embeds = x.data[actors]
        logits = self.proj(actor_embeds)

        # Apply masks from the environment
        if mask.mask is not None and mask.mask.size0() > 0:
            reshaped_masks = torch.tensor(
                mask.mask.as_array().reshape(logits.shape)
            ).to(x.data.device)
            logits = logits.masked_fill(reshaped_masks == 0, -float("inf"))

        dist = Categorical(logits=logits)
        if prev_actions is None:
            action = dist.sample()
        else:
            action = torch.tensor(prev_actions.as_array().squeeze(-1)).to(x.data.device)
        logprob = dist.log_prob(action)
        entropy = dist.entropy()
        return action, lengths, logprob, entropy, dist.logits


def layer_init(
    layer: nn.Module,
    std: float = np.sqrt(2),
    bias_const: float = 0.0,
) -> nn.Module:
    torch.nn.init.orthogonal_(layer.weight, std)
    torch.nn.init.constant_(layer.bias, bias_const)  # type: ignore
    return layer



import math
from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
import torch
from entity_gym.env import VecActionMask, VecSelectEntityActionMask
from ragged_buffer import RaggedBufferI64
from torch import nn
from torch.distributions.categorical import Categorical

from rogue_net.ragged_tensor import RaggedTensor


class PaddedSelectEntityActionHead(nn.Module):
    """
    Action head for selecting entities.
    See https://github.com/entity-neural-network/incubator/pull/109 for more details.
    """

    def __init__(self, d_model: int, d_qk: int) -> None:
        super().__init__()
        self.d_model = d_model
        self.d_qk = d_qk
        self.query_proj = nn.Linear(d_model, d_qk)
        self.key_proj = nn.Linear(d_model, d_qk)

    def forward(
        self,
        x: RaggedTensor,
        index_offsets: RaggedBufferI64,
        mask: VecActionMask,
        prev_actions: Optional[RaggedBufferI64],
    ) -> Tuple[
        torch.Tensor, npt.NDArray[np.int64], torch.Tensor, torch.Tensor, torch.Tensor
    ]:
        assert isinstance(
            mask, VecSelectEntityActionMask
        ), f"Expected SelectEntityActionMaskBatch, got {type(mask)}"
        device = x.data.device
        actor_lengths = mask.actors.size1()
        if len(mask.actors) == 0:
            return (
                torch.zeros((0), dtype=torch.int64, device=device),
                actor_lengths,
                torch.zeros((0), dtype=torch.float32, device=device),
                torch.zeros((0), dtype=torch.float32, device=device),
                torch.zeros((0), dtype=torch.float32, device=device),
            )

        actors = torch.tensor(
            (mask.actors + index_offsets).as_array(), device=device
        ).squeeze(-1)
        actor_embeds = x.data[actors]
        queries = self.query_proj(actor_embeds).squeeze(1)
        max_actors = actor_lengths.max()
        # TODO: can omit rows that are only padding
        padded_queries = torch.zeros(
            len(actor_lengths) * max_actors, self.d_qk, device=device
        )
        qindices = torch.tensor(
            (mask.actors.indices(1) + mask.actors.indices(0) * max_actors)
            .as_array()
            .flatten(),
            device=device,
        )
        padded_queries[qindices] = queries
        padded_queries = padded_queries.view(len(actor_lengths), max_actors, self.d_qk)
        query_mask = torch.zeros(len(actor_lengths) * max_actors, device=device)
        query_mask[qindices] = 1
        query_mask = query_mask.view(len(actor_lengths), max_actors)

        actee_lengths = mask.actees.size1()
        actees = torch.tensor(
            (mask.actees + index_offsets).as_array(), device=device
        ).squeeze(-1)
        actee_embeds = x.data[actees]
        keys = self.key_proj(actee_embeds).squeeze(1)
        max_actees = actee_lengths.max()
        padded_keys = torch.ones(
            len(actee_lengths) * max_actees, self.d_qk, device=device
        )
        kindices = torch.tensor(
            (mask.actees.indices(1) + mask.actees.indices(0) * max_actees)
            .as_array()
            .flatten(),
            device=device,
        )
        padded_keys[kindices] = keys
        padded_keys = padded_keys.view(len(actee_lengths), max_actees, self.d_qk)
        key_mask = torch.zeros(len(actee_lengths) * max_actees, device=device)
        key_mask[kindices] = 1
        key_mask = key_mask.view(len(actee_lengths), max_actees)

        logits = torch.bmm(padded_queries, padded_keys.transpose(1, 2)) * (
            1.0 / math.sqrt(self.d_qk)
        )
        logits_mask = torch.bmm(query_mask.unsqueeze(2), key_mask.unsqueeze(1))

        # Firstly mask off the conditions that are not available. This is the typical masked transformer approach
        logits = logits.masked_fill(logits_mask == 0, -1e9)

        dist = Categorical(logits=logits)
        if prev_actions is None:
            action = dist.sample()
        else:
            action = torch.tensor(prev_actions.as_array(), device=device).flatten()
            padded_actions = torch.ones(
                (logits.size(0) * logits.size(1)), dtype=torch.long, device=device
            )
            padded_actions[qindices] = action
            action = padded_actions.view(len(actor_lengths), max_actors)
        logprob = dist.log_prob(action)
        entropy = dist.entropy()
        return (
            action.flatten()[qindices],
            actor_lengths,
            logprob.flatten()[qindices],
            entropy.flatten()[qindices],
            logits,
        )



import dataclasses
from dataclasses import dataclass
from typing import Dict, Mapping, Optional, Tuple, Type, TypeVar

import numpy as np
import numpy.typing as npt
import torch
import torch.nn as nn
import torch_scatter
from entity_gym.env import ActionSpace, ObsSpace, VecActionMask
from entity_gym.env.environment import (
    CategoricalActionSpace,
    Entity,
    GlobalCategoricalActionSpace,
    SelectEntityActionSpace,
)
from entity_gym.simple_trace import Tracer
from ragged_buffer import (
    RaggedBuffer,
    RaggedBufferBool,
    RaggedBufferF32,
    RaggedBufferI64,
)

from rogue_net.categorical_action_head import CategoricalActionHead
from rogue_net.embedding import EntityEmbedding
from rogue_net.ragged_tensor import RaggedTensor
from rogue_net.select_entity_action_head import PaddedSelectEntityActionHead
from rogue_net.transformer import Transformer, TransformerConfig
from rogue_net.translate_positions import TranslationConfig

ScalarType = TypeVar("ScalarType", bound=np.generic, covariant=True)


def tensor_dict_to_ragged(
    rb_cls: Type[RaggedBuffer[ScalarType]],
    d: Dict[str, torch.Tensor],
    lengths: Dict[str, np.ndarray],
) -> Dict[str, RaggedBuffer[ScalarType]]:
    result = {}
    for k, v in d.items():
        flattened = v.cpu().numpy()
        if flattened.ndim == 1:
            flattened = flattened.reshape(-1, 1)
        result[k] = rb_cls.from_flattened(flattened, lengths[k])
    return result


@dataclass
class RogueNetConfig(TransformerConfig):
    """RogueNet network parameters.

    Attributes:
        d_qk: dimension of keys and queries in select-entity action heads
        translation: settings for transforming all position features to be centered on one entity
    """

    d_qk: int = 16
    translation: Optional[TranslationConfig] = None


class RogueNet(nn.Module):
    def __init__(
        self,
        cfg: RogueNetConfig,
        obs_space: ObsSpace,
        action_space: Dict[str, ActionSpace],
        regression_heads: Optional[Dict[str, int]] = None,
    ):
        super().__init__()

        global_features = obs_space.global_features
        _obs_space = dataclasses.replace(obs_space, global_features=[])
        if len(global_features) > 0:
            _obs_space.entities = {
                label: Entity(entity.features + global_features)
                for label, entity in _obs_space.entities.items()
            }
        if any(
            isinstance(a, GlobalCategoricalActionSpace) for a in action_space.values()
        ):
            _obs_space.entities["__global__"] = Entity(features=global_features)

        self.d_model = cfg.d_model
        self.action_space = action_space
        self.obs_space = obs_space
        self._obs_space = _obs_space
        self.embedding = EntityEmbedding(_obs_space, cfg.translation, cfg.d_model)
        self.backbone = Transformer(cfg, _obs_space)
        self.action_heads = create_action_heads(action_space, cfg.d_model, cfg.d_qk)
        self.auxiliary_heads = (
            nn.ModuleDict(
                {
                    name: regression_head(cfg.d_model, d_out)
                    for name, d_out in regression_heads.items()
                }
            )
            if regression_heads is not None
            else None
        )

    def device(self) -> torch.device:
        return next(self.parameters()).device

    def batch_and_embed(
        self,
        entities: Mapping[str, RaggedBufferF32],
        visible: Mapping[str, RaggedBufferBool],
        tracer: Tracer,
    ) -> RaggedTensor:
        with tracer.span("embedding"):
            # Ensure consistent dictionary ordering
            entities = {
                name: entities[name]
                for name in list(self.obs_space.entities.keys()) + ["__global__"]
                if name in entities
            }
            (
                x,
                tbatch_index,
                index_map,
                tentities,
                tindex_map,
                entity_types,
                tlengths,
            ) = self.embedding(entities, tracer, self.device())

        with tracer.span("visibility_mask"):
            if len(visible) > 0:
                visibilities = []
                for etype, feats in entities.items():
                    if etype in visible:
                        visibilities.append(
                            torch.tensor(
                                visible[etype].as_array(), device=self.device()
                            ).view(-1)
                        )
                    else:
                        visibilities.append(
                            torch.ones(
                                feats.items(),
                                dtype=torch.bool,
                                device=self.device(),
                            )
                        )
                tvisible: Optional[torch.Tensor] = torch.cat(visibilities, dim=0)[
                    tindex_map
                ]
            else:
                tvisible = None

        with tracer.span("backbone"):
            x = self.backbone(
                x,
                tbatch_index,
                index_map,
                tentities,
                tindex_map,
                entity_types,
                tvisible,
            )

        return RaggedTensor(
            x,
            tbatch_index,
            tlengths,
        )

    def get_auxiliary_head(
        self,
        entities: Mapping[str, RaggedBufferF32],
        visible: Mapping[str, RaggedBufferBool],
        head_name: str,
        tracer: Tracer,
    ) -> torch.Tensor:
        x = self.batch_and_embed(entities, visible, tracer)
        pooled = torch_scatter.scatter(
            src=x.data, dim=0, index=x.batch_index, reduce="mean"
        )
        return self.auxiliary_heads[head_name](pooled)  # type: ignore

    def get_action_and_auxiliary(
        self,
        entities: Mapping[str, RaggedBufferF32],
        visible: Mapping[str, RaggedBufferBool],
        action_masks: Mapping[str, VecActionMask],
        tracer: Tracer,
        prev_actions: Optional[Dict[str, RaggedBufferI64]] = None,
    ) -> Tuple[
        Dict[str, RaggedBufferI64],  # actions
        Dict[str, torch.Tensor],  # chosen action probabilities
        Dict[str, torch.Tensor],  # entropy
        Dict[str, npt.NDArray[np.int64]],  # number of actors in each frame
        Dict[str, torch.Tensor],  # auxiliary head values
        Dict[str, torch.Tensor],  # full logits
    ]:
        actions = {}
        probs: Dict[str, torch.Tensor] = {}
        entropies: Dict[str, torch.Tensor] = {}
        logits: Dict[str, torch.Tensor] = {}
        with tracer.span("batch_and_embed"):
            x = self.batch_and_embed(entities, visible, tracer)

        tracer.start("action_heads")
        index_offsets = RaggedBufferI64.from_array(
            torch.cat([torch.tensor([0]).to(self.device()), x.lengths[:-1]])
            .cumsum(0)
            .cpu()
            .numpy()
            .reshape(-1, 1, 1)
        )
        actor_counts: Dict[str, np.ndarray] = {}
        for action_name, action_head in self.action_heads.items():
            action, count, logprob, entropy, logit = action_head(
                x,
                index_offsets,
                action_masks[action_name],
                prev_actions[action_name] if prev_actions is not None else None,
            )
            actor_counts[action_name] = count
            actions[action_name] = action
            probs[action_name] = logprob
            entropies[action_name] = entropy
            if logit is not None:
                logits[action_name] = logit

        tracer.end("action_heads")

        tracer.start("auxiliary_heads")
        if self.auxiliary_heads:
            pooled = torch.zeros(
                x.lengths.size(0), x.data.size(1), device=x.data.device
            )
            torch_scatter.scatter(
                src=x.data,
                dim=0,
                index=x.batch_index,
                reduce="mean",
                out=pooled,
            )
            auxiliary_values = {
                name: module(pooled) for name, module in self.auxiliary_heads.items()
            }
        else:
            auxiliary_values = {}
        tracer.end("auxiliary_heads")

        return (
            prev_actions
            or tensor_dict_to_ragged(RaggedBufferI64, actions, actor_counts),
            probs,
            entropies,
            actor_counts,
            auxiliary_values,
            logits,
        )


def regression_head(d_model: int, d_out: int) -> nn.Module:
    projection = nn.Linear(d_model, d_out)
    projection.weight.data.fill_(0.0)
    projection.bias.data.fill_(0.0)
    return projection


def create_action_heads(
    action_space: Dict[str, ActionSpace], d_model: int, d_qk: int
) -> nn.ModuleDict:
    action_heads: Dict[str, nn.Module] = {}
    for name, space in action_space.items():
        if isinstance(space, CategoricalActionSpace) or isinstance(
            space, GlobalCategoricalActionSpace
        ):
            action_heads[name] = CategoricalActionHead(d_model, len(space))
        elif isinstance(space, SelectEntityActionSpace):
            action_heads[name] = PaddedSelectEntityActionHead(d_model, d_qk)
        else:
            raise ValueError(f"Unknown action space {space}")
    return nn.ModuleDict(action_heads)



import logging
import math
from dataclasses import dataclass
from typing import Literal, Mapping, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch_scatter
from entity_gym.env.environment import ObsSpace
from ragged_buffer import RaggedBufferI64
from torch.nn import functional as F

from rogue_net.relpos_encoding import RelposEncoding, RelposEncodingConfig

logger = logging.getLogger(__name__)


@dataclass
class TransformerConfig:
    """Transformer network hyperparameters.

    Attributes:
        embd_pdrop: Dropout probability for embedding layer.
        resid_pdrop: Dropout probability for residual branches.
        attn_pdrop: Dropout probability for attention.
        n_layer: Number of transformer layers.
        n_head: Number of attention heads.
        d_model: Dimension of embedding.
        pooling: Replace attention with "mean", "max", or "meanmax" pooling.
        relpos_encoding: Relative positional encoding settings.
    """

    embd_pdrop: float = 0.0
    resid_pdrop: float = 0.0
    attn_pdrop: float = 0.0
    n_layer: int = 1
    n_head: int = 2
    d_model: int = 64
    pooling: Optional[Literal["mean", "max", "meanmax"]] = None
    relpos_encoding: Optional[RelposEncodingConfig] = None


class Pool(nn.Module):
    """
    A vanilla multi-head masked self-attention layer with a projection at the end.
    It is possible to use torch.nn.MultiheadAttention here but I am including an
    explicit implementation here to show that there is nothing too scary here.
    """

    def __init__(self, config: TransformerConfig) -> None:
        super().__init__()
        assert config.pooling is not None
        # projections
        self.prepool = nn.Linear(config.d_model, config.d_model)
        if config.pooling == "meanmax":
            self.proj = nn.Linear(2 * config.d_model, config.d_model)
        else:
            self.proj = nn.Linear(config.d_model, config.d_model)
        # regularization
        self.resid_drop = nn.Dropout(config.resid_pdrop)
        self.reduction_op = config.pooling

    def forward(
        self, x: torch.Tensor, batch_index: torch.Tensor, shape: RaggedBufferI64
    ) -> torch.Tensor:
        x = self.prepool(x)

        if "mean" in self.reduction_op:
            xmean = torch_scatter.scatter(
                src=x, dim=0, index=batch_index, reduce="mean"
            )
            xpool = xmean
        if "max" in self.reduction_op:
            xmax = torch_scatter.scatter(src=x, dim=0, index=batch_index, reduce="max")
            xpool = xmax
        if "meanmax" in self.reduction_op:
            xpool = torch.cat([xmean, xmax], dim=1)
        x = self.proj(xpool)

        return self.resid_drop(x[batch_index])  # type: ignore


class RaggedAttention(nn.Module):
    """
    A ragged multi-head masked self-attention layer with a projection at the end.
    It is possible to use torch.nn.MultiheadAttention here but I am including an
    explicit implementation here to show that there is nothing too scary here.
    """

    def __init__(
        self, config: TransformerConfig, relpos_encoding: Optional[RelposEncoding]
    ) -> None:
        super().__init__()
        assert config.d_model % config.n_head == 0
        # key, query, value projections for all heads
        self.key = nn.Linear(config.d_model, config.d_model)
        self.query = nn.Linear(config.d_model, config.d_model)
        self.value = nn.Linear(config.d_model, config.d_model)
        # regularization
        self.attn_drop = nn.Dropout(config.attn_pdrop)
        self.resid_drop = nn.Dropout(config.resid_pdrop)
        # output projection
        self.proj = nn.Linear(config.d_model, config.d_model)
        self.n_head = config.n_head
        self.relpos_encoding = relpos_encoding

    def forward(
        self,
        x: torch.Tensor,
        batch_index: torch.Tensor,
        shape: RaggedBufferI64,
        visible: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # For more details on the implementation, see: https://github.com/entity-neural-network/incubator/pull/119
        device = x.device
        padpack = shape.padpack()

        # TODO: only compute indices once
        if padpack is None:
            nbatch = shape.size0()
            nseq = shape.size1(0) if shape.items() > 0 else 0
            x = x.reshape(nbatch, nseq, x.size(-1))
            if visible is not None:
                attn_mask: Optional[torch.Tensor] = (
                    visible.reshape(nbatch, nseq, 1) > visible.reshape(nbatch, 1, nseq)
                ).unsqueeze(1)
            else:
                attn_mask = None
        else:
            (
                padpack_index,
                padpack_batch,
                padpack_inverse_index,
            ) = padpack
            tpadpack_index = torch.tensor(
                padpack_index, dtype=torch.long, device=device
            )
            x = x[tpadpack_index]
            tpadpack_batch = torch.tensor(padpack_batch, device=device)
            attn_mask = (
                tpadpack_batch.unsqueeze(2) != tpadpack_batch.unsqueeze(1)
            ).unsqueeze(1)
            if visible is not None:
                visible = visible[tpadpack_index]
                attn_mask.logical_or_(
                    (visible.unsqueeze(2) > visible.unsqueeze(1)).unsqueeze(1)
                )

        B, T, C = x.size()

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        k = (
            self.key(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)
        q = (
            self.query(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)
        v = (
            self.value(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)

        # full self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))

        # Relative positional encoding (keys)
        if self.relpos_encoding is not None:
            att += self.relpos_encoding.relattn_logits(q)

        if attn_mask is not None:
            att = att.masked_fill(attn_mask, -1e9)
        att = F.softmax(att, dim=-1)
        att = self.attn_drop(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)

        # Relative positional encoding (values)
        if self.relpos_encoding is not None:
            y += self.relpos_encoding.relpos_values(att, x)

        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_drop(self.proj(y))
        if padpack is None:
            return y.reshape(batch_index.size(0), y.size(-1))  # type: ignore
        else:
            return y.reshape(y.size(0) * y.size(1), y.size(2))[torch.tensor(padpack_inverse_index, dtype=torch.long, device=device)]  # type: ignore


class Block(nn.Module):
    def __init__(
        self, config: TransformerConfig, relpos_encoding: Optional[RelposEncoding]
    ) -> None:
        super().__init__()
        self.ln1 = nn.LayerNorm(config.d_model)
        self.ln2 = nn.LayerNorm(config.d_model)
        if config.pooling is not None:
            self.attn: Union[Pool, RaggedAttention] = Pool(config)
        else:
            self.attn = RaggedAttention(config, relpos_encoding)
        self.mlp = nn.Sequential(
            nn.Linear(config.d_model, 4 * config.d_model),
            nn.GELU(),
            nn.Linear(4 * config.d_model, config.d_model),
            nn.Dropout(config.resid_pdrop),
        )

    def forward(
        self,
        x: torch.Tensor,
        batch_index: torch.Tensor,
        shape: RaggedBufferI64,
        visible: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        x = x + self.attn(self.ln1(x), batch_index, shape, visible)
        x = x + self.mlp(self.ln2(x))
        return x


class Transformer(nn.Module):
    def __init__(self, config: TransformerConfig, obs_space: ObsSpace) -> None:
        super().__init__()

        if config.relpos_encoding is not None:
            self.relpos_encoding: Optional[RelposEncoding] = RelposEncoding(
                config.relpos_encoding,
                obs_space,
                dmodel=config.d_model,
                dhead=config.d_model // config.n_head,
            )
        else:
            self.relpos_encoding = None
        self.drop = nn.Dropout(config.embd_pdrop)
        self.blocks = nn.Sequential(
            *[Block(config, self.relpos_encoding) for _ in range(config.n_layer)]
        )

        self.apply(self._init_weights)

        logger.info(
            "number of parameters: %e", sum(p.numel() for p in self.parameters())
        )

    def _init_weights(self, module: nn.Module) -> None:
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self,
        x: torch.Tensor,
        batch_index: torch.Tensor,
        shape: RaggedBufferI64,
        input_feats: Mapping[str, torch.Tensor],
        index_map: torch.Tensor,
        entity_type: torch.Tensor,
        visible: Optional[torch.Tensor],
    ) -> torch.Tensor:
        x = self.drop(x)

        if self.relpos_encoding is not None:
            device = x.device
            padpack = shape.padpack()
            if padpack is None:
                tpadpack_index = None
            else:
                tpadpack_index = torch.tensor(
                    padpack[0], dtype=torch.long, device=device
                )
            relkeysvals: Optional[
                Tuple[torch.Tensor, torch.Tensor]
            ] = self.relpos_encoding.keys_values(
                input_feats,
                index_map,
                tpadpack_index,
                shape,
                entity_type,
            )
            self.relpos_encoding.cached_rkvs = relkeysvals
        else:
            relkeysvals = None

        for block in self.blocks:
            x = block(x, batch_index, shape, visible)
        return x



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rogue_net', 'rogue_net.tests']

package_data = \
{'': ['*']}

install_requires = \
['entity-gym>=0.1.2,<0.2.0', 'ragged-buffer>=0.3.9,<0.4.0']

setup_kwargs = {
    'name': 'rogue-net',
    'version': '0.1.1',
    'description': 'Ragged batch transformer implementation that is compatible with entity-gym',
    'long_description': None,
    'author': 'Clemens Winter',
    'author_email': 'clemenswinter1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Dict, Mapping, Optional, Tuple

import numpy as np
import ragged_buffer
import torch
from entity_gym.env import ObsSpace
from entity_gym.simple_trace import Tracer
from ragged_buffer import RaggedBufferF32, RaggedBufferI64
from torch import nn

from rogue_net.input_norm import InputNorm
from rogue_net.translate_positions import TranslatePositions, TranslationConfig


class EntityEmbedding(nn.Module):
    def __init__(
        self,
        obs_space: ObsSpace,
        feature_transforms: Optional[TranslationConfig],
        d_model: int,
    ) -> None:
        super().__init__()
        if feature_transforms is not None:
            self.feature_transforms: Optional[TranslatePositions] = TranslatePositions(
                feature_transforms, obs_space
            )
            obs_space = self.feature_transforms.transform_obs_space(obs_space)
        else:
            self.feature_transforms = None

        embeddings: Dict[str, nn.Module] = {}
        for name, entity in obs_space.entities.items():
            if entity.features:
                embeddings[name] = nn.Sequential(
                    InputNorm(len(entity.features)),
                    nn.Linear(len(entity.features), d_model),
                    nn.ReLU(),
                    nn.LayerNorm(d_model),
                )
            else:
                embeddings[name] = FeaturelessEmbedding(d_model)
        self.embeddings = nn.ModuleDict(embeddings)

    def forward(
        self,
        entities: Mapping[str, RaggedBufferF32],
        tracer: Tracer,
        device: torch.device,
    ) -> Tuple[
        torch.Tensor,
        torch.Tensor,
        RaggedBufferI64,
        Mapping[str, torch.Tensor],
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
    ]:
        entity_embeds = []
        index_offsets = {}
        index_offset = 0
        entity_type = []

        if "__global__" in entities:
            globals = entities["__global__"]
            entities = {
                label: ragged_buffer.cat([feats, globals], dim=2)
                if label != "__global__"
                else feats
                for label, feats in entities.items()
            }

        if self.feature_transforms:
            entities = {name: feats.clone() for name, feats in entities.items()}
            self.feature_transforms.apply(entities)
        tentities = {
            entity: torch.tensor(features.as_array(), device=device)
            for entity, features in entities.items()
        }

        for i, (entity, embedding) in enumerate(self.embeddings.items()):
            # We may have environment states that do not contain every possible entity
            if entity in entities:
                batch = tentities[entity]
                emb = embedding(batch)
                entity_embeds.append(emb)
                entity_type.append(
                    torch.full((emb.size(0), 1), float(i), device=device)
                )
                index_offsets[entity] = index_offset
                index_offset += batch.size(0)

        x = torch.cat(entity_embeds)
        with tracer.span("ragged_metadata"):
            real_entities = {name: entities[name] for name in self.embeddings.keys()}
            lengths = sum(entity.size1() for entity in real_entities.values())
            batch_index = np.concatenate(
                [
                    entity.indices(0).as_array().flatten()
                    for entity in real_entities.values()
                ]
            )
            index_map = ragged_buffer.cat(
                [
                    entity.flat_indices() + index_offsets[name]
                    for name, entity in real_entities.items()
                    if name in self.embeddings
                ],
                dim=1,
            )
            tindex_map = torch.tensor(index_map.as_array().flatten(), device=device)
            tbatch_index = torch.tensor(batch_index, device=device)
            tlengths = torch.tensor(lengths, device=device)

        x = x[tindex_map]
        entity_types = torch.cat(entity_type)[tindex_map]
        tbatch_index = tbatch_index[tindex_map]

        return (
            x,
            tbatch_index,
            index_map,
            tentities,
            tindex_map,
            entity_types,
            tlengths,
        )


class FeaturelessEmbedding(nn.Module):
    def __init__(self, d_model: int):
        super().__init__()
        self.d_model = d_model
        self.embedding = nn.Parameter(torch.randn(1, d_model))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embedding.repeat(x.size(0), 1)
